@@ -24,6 +24,7 @@
 #include <sys/types.h>
 #include <unistd.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 #include <netdb.h> 
 #include <netinet/in.h>
 #include <X11/Xauth.h>
@@ -37,7 +38,7 @@
 #include "auth.h"
 
 /* Local prototypes */
-static void gdm_auth_purge (GdmDisplay *d, FILE *af);
+static FILE *gdm_auth_purge (GdmDisplay *d, FILE *af);
 
 /* Configuration option variables */
 extern gchar *GdmServAuthDir;
@@ -46,7 +47,6 @@ extern gchar *GdmUserAuthFile;
 extern gchar *GdmUserAuthFB;
 extern gint  GdmUserMaxFile;
 extern gint  GdmRelaxPerms;
-extern gboolean GdmDebug;
 
 static void
 display_add_error (GdmDisplay *d)
@@ -76,25 +76,30 @@ add_auth_entry (GdmDisplay *d, FILE *af, FILE *af2,
 	Xauth *xa;
 	gchar *dispnum;
 
-	if (!d)
+	if G_UNLIKELY (!d)
 		return FALSE;
 
 	xa = malloc (sizeof (Xauth));
 
-	if (xa == NULL)
+	if G_UNLIKELY (xa == NULL)
 		return FALSE;
 
 	xa->family = family;
-	xa->address = malloc (addrlen);
-	if (xa->address == NULL) {
-		free (xa);
-		return FALSE;
+	if (addrlen == 0) {
+		xa->address = NULL;
+		xa->address_length = 0;
+	} else {
+		xa->address = malloc (addrlen);
+		if G_UNLIKELY (xa->address == NULL) {
+			free (xa);
+			return FALSE;
+		}
+
+
+		memcpy (xa->address, addr, addrlen);
+		xa->address_length = addrlen;
 	}
 
-
-	memcpy (xa->address, addr, addrlen);
-	xa->address_length = addrlen;
-
 	dispnum = g_strdup_printf ("%d", d->dispnum);
 	xa->number = strdup (dispnum);
 	xa->number_length = strlen (dispnum);
@@ -103,7 +108,7 @@ add_auth_entry (GdmDisplay *d, FILE *af, FILE *af2,
 	xa->name = strdup ("MIT-MAGIC-COOKIE-1");
 	xa->name_length = strlen ("MIT-MAGIC-COOKIE-1");
 	xa->data = malloc (16);
-	if (xa->data == NULL) {
+	if G_UNLIKELY (xa->data == NULL) {
 		free (xa->number);
 		free (xa->name);
 		free (xa->address);
@@ -114,7 +119,7 @@ add_auth_entry (GdmDisplay *d, FILE *af, FILE *af2,
 	xa->data_length = 16;
 
 	errno = 0;
-	if ( ! XauWriteAuth (af, xa)) {
+	if G_UNLIKELY ( ! XauWriteAuth (af, xa)) {
 		free (xa->data);
 		free (xa->number);
 		free (xa->name);
@@ -126,7 +131,7 @@ add_auth_entry (GdmDisplay *d, FILE *af, FILE *af2,
 
 	if (af2 != NULL) {
 		errno = 0;
-		if ( ! XauWriteAuth (af2, xa)) {
+		if G_UNLIKELY ( ! XauWriteAuth (af2, xa)) {
 			free (xa->data);
 			free (xa->number);
 			free (xa->name);
@@ -159,8 +164,9 @@ gdm_auth_secure_display (GdmDisplay *d)
     const char lo[] = {127,0,0,1};
     guint i;
     const GList *local_addys = NULL;
+    gboolean added_lo = FALSE;
 
-    if (!d)
+    if G_UNLIKELY (!d)
 	return FALSE;
 
     umask (022);
@@ -178,13 +184,13 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    /* Note, Xnest can't use the ServAuthDir unless running as
 	     * root, which is rare anyway, unless the user is a wanker */
 
-	    d->authfile = g_strconcat (GdmUserAuthFB, "/.gdmXXXXXX", NULL);
+	    d->authfile = g_build_filename (GdmUserAuthFB, ".gdmXXXXXX", NULL);
 
 	    umask (077);
 	    authfd = g_mkstemp (d->authfile);
 	    umask (022);
 
-	    if (authfd == -1) {
+	    if G_UNLIKELY (authfd == -1) {
 		    gdm_error (_("%s: Could not make new cookie file in %s"),
 			       "gdm_auth_secure_display", GdmUserAuthFB);
 		    g_free (d->authfile);
@@ -197,7 +203,7 @@ gdm_auth_secure_display (GdmDisplay *d)
 
 	    af = fdopen (authfd, "w");
 
-	    if (af == NULL) {
+	    if G_UNLIKELY (af == NULL) {
 		    g_free (d->authfile);
 		    d->authfile = NULL;
 		    return FALSE;
@@ -208,7 +214,7 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    d->authfile_gdm = g_strconcat (GdmServAuthDir, "/", d->name, ".Xauth", NULL);
 	    af_gdm = gdm_safe_fopen_w (d->authfile_gdm);
 
-	    if (af_gdm == NULL) {
+	    if G_UNLIKELY (af_gdm == NULL) {
 		    gdm_error (_("%s: Cannot safely open %s"),
 			       "gdm_auth_user_remove",
 			       d->authfile_gdm);
@@ -225,7 +231,7 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    d->authfile = g_strconcat (GdmServAuthDir, "/", d->name, ".Xauth", NULL);
 	    af = gdm_safe_fopen_w (d->authfile);
 
-	    if (af == NULL) {
+	    if G_UNLIKELY (af == NULL) {
 		    gdm_error (_("%s: Cannot safely open %s"),
 			       "gdm_auth_user_remove",
 			       d->authfile);
@@ -266,7 +272,7 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    char hostname[1024];
 
 	    hostname[1023] = '\0';
-	    if (gethostname (hostname, 1023) == 0) {
+	    if G_LIKELY (gethostname (hostname, 1023) == 0) {
 		    g_free (d->hostname);
 		    d->hostname = g_strdup (hostname);
 	    }
@@ -329,6 +335,9 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    if ( ! add_auth_entry (d, af, af_gdm, FamilyInternet,
 				   (char *)&(d->addr), 4))
 		    return FALSE;
+
+	    if (gdm_is_loopback_addr (&(d->addr)))
+		    added_lo = TRUE;
     }
     
     /* Network access: Write out an authentication entry for each of
@@ -342,6 +351,9 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    if ( ! add_auth_entry (d, af, af_gdm, FamilyInternet,
 				   (char *)&ia->s_addr, 4))
 		    return FALSE;
+
+	    if (gdm_is_loopback_addr (ia))
+		    added_lo = TRUE;
     }
 
     /* Network access: Write out an authentication entry for each of
@@ -355,33 +367,44 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    if ( ! add_auth_entry (d, af, af_gdm, FamilyInternet,
 				   (char *)&ia->s_addr, 4))
 		    return FALSE;
+
+	    if (gdm_is_loopback_addr (ia))
+		    added_lo = TRUE;
     }
 
     /* if local server add loopback */
-    if (SERVER_IS_LOCAL (d)) {
+    if (SERVER_IS_LOCAL (d) && ! added_lo) {
 	    if ( ! add_auth_entry (d, af, af_gdm, FamilyInternet,
 				   lo, 4))
 		    return FALSE;
     }
 
-    fclose (af);
-    if (af_gdm != NULL)
-	    fclose (af_gdm);
-    ve_setenv ("XAUTHORITY", d->authfile, TRUE);
+    if G_UNLIKELY (fclose (af) < 0) {
+	    display_add_error (d);
+	    return FALSE;
+    }
+    if (af_gdm != NULL) {
+	    if G_UNLIKELY (fclose (af_gdm) < 0) {
+		    display_add_error (d);
+		    return FALSE;
+	    }
+    }
+    ve_setenv ("XAUTHORITY", GDM_AUTHFILE (d), TRUE);
 
-    if (GdmDebug)
-	    gdm_debug ("gdm_auth_secure_display: Setting up access for %s - %d entries", 
-		       d->name, g_slist_length (d->auths));
+    gdm_debug ("gdm_auth_secure_display: Setting up access for %s - %d entries", 
+	       d->name, g_slist_length (d->auths));
 
     return TRUE;
 }
 
+
 static gboolean
 try_open_append (const char *file)
 {
 	FILE *fp;
+
 	fp = fopen (file, "a+");
-	if (fp != NULL) {
+	if G_LIKELY (fp != NULL) {
 		fclose (fp);
 		return TRUE;
 	} else {
@@ -389,6 +412,28 @@ try_open_append (const char *file)
 	}
 }
 
+static gboolean
+try_open_read_as_root (const char *file)
+{
+	int fd;
+	uid_t oldeuid = geteuid ();
+	uid_t oldegid = getegid ();
+	setegid (0);
+	seteuid (0);
+
+	fd = open (file, O_RDONLY);
+	if G_UNLIKELY (fd < 0) {
+		seteuid (oldeuid);
+		setegid (oldegid);
+		return FALSE;
+	} else {
+		close (fd);
+		seteuid (oldeuid);
+		setegid (oldegid);
+		return TRUE;
+	}
+}
+
 /**
  * gdm_auth_user_add:
  * @d: Pointer to a GdmDisplay struct
@@ -410,6 +455,7 @@ gdm_auth_user_add (GdmDisplay *d, uid_t user, const char *homedir)
     GSList *auths = NULL;
     gboolean ret = TRUE;
     gboolean automatic_tmp_dir = FALSE;
+    gboolean authdir_is_tmp_dir = FALSE;
     gboolean locked;
 
     if (!d)
@@ -424,6 +470,8 @@ gdm_auth_user_add (GdmDisplay *d, uid_t user, const char *homedir)
 		    authdir = g_strconcat (homedir, &GdmUserAuthDir[1], NULL);
 	    } else {
 		    authdir = g_strdup (GdmUserAuthDir);
+		    automatic_tmp_dir = TRUE;
+		    authdir_is_tmp_dir = TRUE;
 	    }
     } else {
 	    authdir = g_strdup (homedir);
@@ -438,22 +486,41 @@ try_user_add_again:
     if (authdir == NULL)
 	    d->userauth = NULL;
     else
-	    d->userauth = g_strconcat (authdir, "/", GdmUserAuthFile, NULL);
+	    d->userauth = g_build_filename (authdir, GdmUserAuthFile, NULL);
 
     /* Find out if the Xauthority file passes the paranoia check */
     if (automatic_tmp_dir ||
 	authdir == NULL ||
 	! gdm_file_check ("gdm_auth_user_add", user, authdir, GdmUserAuthFile, 
 			  TRUE, FALSE, GdmUserMaxFile, GdmRelaxPerms) ||
-	! try_open_append (d->userauth)) {
+	! try_open_append (d->userauth) ||
 
-	/* No go. Let's create a fallback file in GdmUserAuthFB (/tmp) */
+	/* try opening as root, if we can't open as root,
+	   then this is a NFS mounted directory with root squashing,
+	   and we don't want to write cookies over NFS */
+	! try_open_read_as_root (d->userauth)) {
+
+	/* No go. Let's create a fallback file in GdmUserAuthFB (/tmp)
+	 * or perhaps GdmUserAuth directory (usually would be /tmp) */
 	d->authfb = TRUE;
 	g_free (d->userauth);
-	d->userauth = g_strconcat (GdmUserAuthFB, "/.gdmXXXXXX", NULL);
+	if (authdir_is_tmp_dir && authdir != NULL)
+		d->userauth = g_build_filename (authdir, ".gdmXXXXXX", NULL);
+	else
+		d->userauth = g_build_filename (GdmUserAuthFB, ".gdmXXXXXX", NULL);
 	authfd = g_mkstemp (d->userauth);
 
-	if (authfd == -1) {
+	if G_UNLIKELY (authfd < 0 && authdir_is_tmp_dir) {
+	    g_free (d->userauth);
+	    d->userauth = NULL;
+
+	    umask (022);
+
+	    authdir_is_tmp_dir = FALSE;
+	    goto try_user_add_again;
+	}
+
+	if G_UNLIKELY (authfd < 0) {
 	    gdm_error (_("%s: Could not open cookie file %s"),
 		       "gdm_auth_user_add",
 		       d->userauth);
@@ -472,7 +539,7 @@ try_user_add_again:
 	d->authfb = FALSE;
 
 	/* FIXME: Better implement my own locking. The libXau one is not kosher */
-	if (XauLockAuth (d->userauth, 3, 3, 0) != LOCK_SUCCESS) {
+	if G_UNLIKELY (XauLockAuth (d->userauth, 3, 3, 0) != LOCK_SUCCESS) {
 	    gdm_error (_("%s: Could not lock cookie file %s"),
 		       "gdm_auth_user_add",
 		       d->userauth);
@@ -490,7 +557,7 @@ try_user_add_again:
 	af = gdm_safe_fopen_ap (d->userauth);
     }
 
-    if (!af) {
+    if G_UNLIKELY (af == NULL) {
 	/* Really no need to clean up here - this process is a goner anyway */
 	gdm_error (_("%s: Could not open cookie file %s"),
 		   "gdm_auth_user_add",
@@ -516,13 +583,13 @@ try_user_add_again:
 
     /* If not a fallback file, nuke any existing cookies for this display */
     if (! d->authfb)
-	gdm_auth_purge (d, af);
+	af = gdm_auth_purge (d, af);
 
     /* Append the authlist for this display to the cookie file */
     auths = d->auths;
 
     while (auths) {
-	if ( ! XauWriteAuth (af, auths->data)) {
+	if G_UNLIKELY ( ! XauWriteAuth (af, auths->data)) {
 		gdm_error (_("%s: Could not write cookie"),
 			   "gdm_auth_user_add");
 
@@ -543,7 +610,22 @@ try_user_add_again:
 	auths = auths->next;
     }
 
-    fclose (af);
+    if G_UNLIKELY (fclose (af) < 0) {
+	    gdm_error (_("%s: Could not write cookie"),
+		       "gdm_auth_user_add");
+
+	    if ( ! d->authfb) {
+		    if (locked)
+			    XauUnlockAuth (d->userauth);
+		    g_free (d->userauth);
+		    d->userauth = NULL;
+		    automatic_tmp_dir = TRUE;
+		    goto try_user_add_again;
+	    }
+
+	    ret = FALSE;
+    }
+
     if (locked)
 	    XauUnlockAuth (d->userauth);
 
@@ -572,7 +654,7 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
     gchar *authfile;
     gchar *authdir;
 
-    if (!d || !d->userauth)
+    if G_UNLIKELY (!d || !d->userauth)
 	return;
 
     gdm_debug ("gdm_auth_user_remove: Removing cookie from %s (%d)", d->userauth, d->authfb);
@@ -580,7 +662,14 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
     /* If we are using the fallback cookie location, simply nuke the
      * cookie file */
     if (d->authfb) {
-	unlink (d->userauth);
+	IGNORE_EINTR (unlink (d->userauth));
+	g_free (d->userauth);
+	d->userauth = NULL;
+	return;
+    }
+
+    /* if the file doesn't exist, oh well, just ignore this then */
+    if G_UNLIKELY (access (d->userauth, F_OK) != 0) {
 	g_free (d->userauth);
 	d->userauth = NULL;
 	return;
@@ -593,7 +682,7 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
      * decided to concatenate something like his entire MP3 collection
      * to it. So we better play it safe... */
 
-    if ( ! gdm_file_check ("gdm_auth_user_remove", user, authdir, authfile, 
+    if G_UNLIKELY ( ! gdm_file_check ("gdm_auth_user_remove", user, authdir, authfile, 
 			   TRUE, FALSE, GdmUserMaxFile, GdmRelaxPerms)) {
 	    g_free (authdir);
 	    g_free (authfile);
@@ -608,7 +697,7 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
     g_free (authfile);
 
     /* Lock user's cookie jar and open it for writing */
-    if (XauLockAuth (d->userauth, 3, 3, 0) != LOCK_SUCCESS) {
+    if G_UNLIKELY (XauLockAuth (d->userauth, 3, 3, 0) != LOCK_SUCCESS) {
 	g_free (d->userauth);
 	d->userauth = NULL;
 	return;
@@ -616,7 +705,7 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
 
     af = gdm_safe_fopen_ap (d->userauth);
 
-    if (!af) {
+    if G_UNLIKELY (af == NULL) {
 	XauUnlockAuth (d->userauth);
 
 	gdm_error (_("%s: Cannot safely open %s"),
@@ -630,10 +719,11 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
     }
 
     /* Purge entries for this display from the cookie jar */
-    gdm_auth_purge (d, af);
+    af = gdm_auth_purge (d, af);
 
     /* Close the file and unlock it */
     fclose (af);
+    /* FIXME: what about out of diskspace errors on errors close */
     XauUnlockAuth (d->userauth);
 
     g_free (d->userauth);
@@ -649,14 +739,16 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
  * Remove all cookies referring to this display a cookie file.
  */
 
-static void
+static FILE *
 gdm_auth_purge (GdmDisplay *d, FILE *af)
 {
     Xauth *xa;
     GSList *keep = NULL, *li;
+    char *dispnum;
+    int displen;
 
-    if (!d || !af)
-	return;
+    if G_UNLIKELY (!d || !af)
+	return af;
 
     gdm_debug ("gdm_auth_purge: %s", d->name);
 
@@ -666,41 +758,47 @@ gdm_auth_purge (GdmDisplay *d, FILE *af)
      * temporary file issues. Then remove any instance of this display
      * in the cookie jar... */
 
+    dispnum = g_strdup_printf ("%d", d->dispnum);
+    displen = strlen (dispnum);
+
     while ( (xa = XauReadAuth (af)) != NULL ) {
-	gboolean match = FALSE;
-	GSList *alist = d->auths;
 
-	while (alist) {
-	    Xauth *da = alist->data;
-
-	    if (! memcmp (da->address, xa->address, xa->address_length) &&
-		! memcmp (da->number, xa->number, xa->number_length))
-		match = TRUE;
-
-	    alist = alist->next;
-	}
-
-	if (match)
-	    XauDisposeAuth (xa);
+	if (xa->number_length == displen &&
+	    xa->name_length == strlen ("MIT-MAGIC-COOKIE-1") &&
+	    memcmp ("MIT-MAGIC-COOKIE-1", xa->name, xa->name_length) == 0 &&
+	    memcmp (dispnum, xa->number, xa->number_length) == 0)
+		XauDisposeAuth (xa);
 	else
-	    keep = g_slist_append (keep, xa);
+		keep = g_slist_append (keep, xa);
     }
 
+    g_free (dispnum);
+
     /* Rewind the file */
-    af = freopen (d->userauth, "w", af);
+    fclose (af);
+    af = gdm_safe_fopen_w (d->userauth);
 
     /* Write out remaining entries */
     for (li = keep; li != NULL; li = li->next) {
 	    /* FIXME: is this correct, if we can't open
 	     * this is quite crap isn't it ... */
-	    if (af != NULL)
+	    if G_LIKELY (af != NULL)
 		    XauWriteAuth (af, li->data);
+	            /* FIXME: what about errors? */
 	    XauDisposeAuth (li->data);
 	    li->data = NULL;
     }
 
     g_slist_free (keep);
+
+    return af;
 }
 
+void
+gdm_auth_set_local_auth (GdmDisplay *d)
+{
+	XSetAuthorization ((char *)"MIT-MAGIC-COOKIE-1", (int) strlen ("MIT-MAGIC-COOKIE-1"),
+			   (char *)d->bcookie, (int) 16);
+}
 
 /* EOF */
