@@ -439,8 +439,9 @@ gdm_display_dispose (GdmDisplay *d)
     g_free (d->authfile_gdm);
     d->authfile_gdm = NULL;
 
-    if (d->xnest_temp_auth_file != NULL)
+    if (d->xnest_temp_auth_file != NULL) {
 	    IGNORE_EINTR (unlink (d->xnest_temp_auth_file));
+    }
     g_free (d->xnest_temp_auth_file);
     d->xnest_temp_auth_file = NULL;
 
