@@ -195,12 +195,12 @@ enum {
 #define GDM_KEY_XDMCP "xdmcp/Enable=false"
 #define GDM_KEY_MAXPEND "xdmcp/MaxPending=4"
 #define GDM_KEY_MAXSESS "xdmcp/MaxSessions=16"
-#define GDM_KEY_MAXWAIT "xdmcp/MaxWait=30"
+#define GDM_KEY_MAXWAIT "xdmcp/MaxWait=15"
 #define GDM_KEY_DISPERHOST "xdmcp/DisplaysPerHost=2"
 #define GDM_KEY_UDPPORT "xdmcp/Port=177"
 #define GDM_KEY_INDIRECT "xdmcp/HonorIndirect=true"
 #define GDM_KEY_MAXINDIR "xdmcp/MaxPendingIndirect=4"
-#define GDM_KEY_MAXINDWAIT "xdmcp/MaxWaitIndirect=30"
+#define GDM_KEY_MAXINDWAIT "xdmcp/MaxWaitIndirect=15"
 #define GDM_KEY_PINGINTERVAL "xdmcp/PingIntervalSeconds=15"
 #define GDM_KEY_WILLING "xdmcp/Willing=" EXPANDED_SYSCONFDIR "/gdm/Xwilling"
 
@@ -211,7 +211,7 @@ enum {
 #define GDM_KEY_BROWSER "greeter/Browser=false"
 #define GDM_KEY_EXCLUDE "greeter/Exclude=bin,daemon,adm,lp,sync,shutdown,halt,mail,news,uucp,operator,nobody,gdm,postgres,pvm,rpm,nfsnobody,pcap"
 #define GDM_KEY_MINIMALUID "greeter/MinimalUID=100"
-#define GDM_KEY_FACE "greeter/DefaultFace=" EXPANDED_PIXMAPDIR "nobody.png"
+#define GDM_KEY_FACE "greeter/DefaultFace=" EXPANDED_PIXMAPDIR "/nobody.png"
 #define GDM_KEY_FACEDIR "greeter/GlobalFaceDir=" EXPANDED_DATADIR "/faces/"
 #define GDM_KEY_LOCFILE "greeter/LocaleFile=" EXPANDED_LOCALEDIR "/locale.alias"
 #define GDM_KEY_LOGO "greeter/Logo=" EXPANDED_PIXMAPDIR "/gdm-foot-logo.png"
@@ -306,6 +306,7 @@ struct _GdmDisplay {
     pid_t greetpid;
     pid_t servpid;
     pid_t sesspid;
+    int last_sess_status; /* status returned by last session */
     pid_t slavepid;
     pid_t chooserpid;
     time_t acctime;
@@ -356,6 +357,13 @@ struct _GdmDisplay {
 
     /* order in the Xservers file for sessreg, -1 if unset yet */
     int x_servers_order;
+
+    /* The xsession-errors connection */
+    int xsession_errors_fd; /* write to the file */
+    int session_output_fd; /* read from the session */
+    int xsession_errors_bytes;
+#define MAX_XSESSION_ERRORS_BYTES (80*2500)  /* maximum number of bytes in
+						the ~/.xsession-errors file */
 };
 
 typedef struct _GdmXServer GdmXServer;
@@ -675,7 +683,7 @@ enum {
 	do {		\
 		errno = 0;	\
 		expr;		\
-	} while (errno == EINTR);
+	} while G_UNLIKELY (errno == EINTR);
 		
 
 #endif /* GDM_H */
