@@ -19,6 +19,10 @@
 #ifndef GDM_MISC_H
 #define GDM_MISC_H
 
+#ifndef __PRETTY_FUNCTION__
+#define __PRETTY_FUNCTION__ "N/A"
+#endif
+
 #include <sys/types.h>
 
 #include "gdm.h"
@@ -28,6 +32,20 @@ void gdm_info   (const gchar *format, ...) G_GNUC_PRINTF (1, 2);
 void gdm_error  (const gchar *format, ...) G_GNUC_PRINTF (1, 2);
 void gdm_debug  (const gchar *format, ...) G_GNUC_PRINTF (1, 2);
 
+#define gdm_assert(expr)		G_STMT_START{			\
+     if G_LIKELY(expr) { } else 					\
+        gdm_fail ("GDM file %s: line %d (%s): assertion failed: (%s)",	\
+		  __FILE__,						\
+		  __LINE__,						\
+		  __PRETTY_FUNCTION__,					\
+                  #expr);			}G_STMT_END
+
+#define gdm_assert_not_reached()	G_STMT_START{			\
+     gdm_fail ("GDM file %s: line %d (%s): should not be reached",	\
+	       __FILE__,						\
+	       __LINE__,						\
+	       __PRETTY_FUNCTION__);	}G_STMT_END
+
 void gdm_fdprintf  (int fd, const gchar *format, ...) G_GNUC_PRINTF (2, 3);
 int gdm_fdgetc     (int fd);
 char *gdm_fdgets   (int fd);
@@ -106,6 +124,9 @@ FILE * gdm_safe_fopen_w (const char *file);
 /* like fopen with "a+" and uses O_EXCL and O_NOFOLLOW */
 FILE * gdm_safe_fopen_ap (const char *file);
 
+/* first must get initial limits before attempting to ever reset those
+   limits */
+void gdm_get_initial_limits (void);
 void gdm_reset_limits (void);
 
 const char *gdm_root_user (void);
