@@ -54,7 +54,6 @@
 /* Local prototypes */
 static void gdm_server_spawn (GdmDisplay *d, const char *vtarg);
 static void gdm_server_usr1_handler (gint);
-static void gdm_server_alarm_handler (gint);
 static void gdm_server_child_handler (gint);
 static char * get_font_path (const char *display);
 
@@ -75,6 +74,9 @@ static GdmDisplay *d = NULL;
 static gboolean server_signal_notified = FALSE;
 static int server_signal_pipe[2];
 
+static void do_server_wait (GdmDisplay *d);
+static gboolean setup_server_wait (GdmDisplay *d);
+
 void
 gdm_server_whack_lockfile (GdmDisplay *disp)
 {
@@ -94,12 +96,12 @@ gdm_server_whack_lockfile (GdmDisplay *disp)
 
 	    /* if lock file exists and it is our process, whack it! */
 	    g_snprintf (buf, sizeof (buf), "/tmp/.X%d-lock", disp->dispnum);
-	    unlink (buf);
+	    IGNORE_EINTR (unlink (buf));
 
 	    /* whack the unix socket as well */
 	    g_snprintf (buf, sizeof (buf),
 			"/tmp/.X11-unix/X%d", disp->dispnum);
-	    unlink (buf);
+	    IGNORE_EINTR (unlink (buf));
 }
 
 
@@ -107,16 +109,20 @@ gdm_server_whack_lockfile (GdmDisplay *disp)
 void
 gdm_server_wipe_cookies (GdmDisplay *disp)
 {
-	if ( ! ve_string_empty (disp->authfile))
-		unlink (disp->authfile);
+	if ( ! ve_string_empty (disp->authfile)) {
+		IGNORE_EINTR (unlink (disp->authfile));
+	}
 	g_free (disp->authfile);
 	disp->authfile = NULL;
-	if ( ! ve_string_empty (disp->authfile_gdm))
-		unlink (disp->authfile_gdm);
+	if ( ! ve_string_empty (disp->authfile_gdm)) {
+		IGNORE_EINTR (unlink (disp->authfile_gdm));
+	}
 	g_free (disp->authfile_gdm);
 	disp->authfile_gdm = NULL;
 }
 
+static Jmp_buf reinitjmp;
+
 /* ignore handlers */
 static int
 ignore_xerror_handler (Display *disp, XErrorEvent *evt)
@@ -125,11 +131,37 @@ ignore_xerror_handler (Display *disp, XErrorEvent *evt)
 }
 
 static int
-exit_xioerror_handler (Display *disp)
+jumpback_xioerror_handler (Display *disp)
 {
-	_exit (1);
+	Longjmp (reinitjmp, 1);
 }
 
+#ifdef HAVE_FBCONSOLE
+#define FBCONSOLE "/usr/openwin/bin/fbconsole"
+
+static void
+gdm_exec_fbconsole(GdmDisplay *disp)
+{
+        pid_t pid;
+        char *argv[6];
+
+        argv[0] = FBCONSOLE;
+        argv[1] = "-d";
+        argv[2] = disp->name;
+        argv[3] = NULL;
+
+        pid = fork ();
+        if (pid == 0) {
+                gdm_close_all_descriptors (0 /* from */, -1 /* except */, -1 /* except2 */)
+;
+                IGNORE_EINTR (execv (argv[0], argv));
+        }
+        if (pid == -1) {
+                gdm_error (_("Can not start fallback console"));
+        }
+}
+#endif
+
 /**
  * gdm_server_reinit:
  * @disp: Pointer to a GdmDisplay structure
@@ -138,88 +170,86 @@ exit_xioerror_handler (Display *disp)
  * but only if the display exists
  */
 
-void
+gboolean
 gdm_server_reinit (GdmDisplay *disp)
 {
-	pid_t pid;
-
 	if (disp == NULL)
-		return;
+		return FALSE;
 
-	/* Kill our connection if one existed */
-	if (disp->dsp != NULL) {
-		XCloseDisplay (disp->dsp);
-		disp->dsp = NULL;
+	if (disp->servpid <= 0) {
+		/* Kill our connection if one existed, likely to result
+		 * in some bizzaro error right now */
+		if (disp->dsp != NULL) {
+			XCloseDisplay (disp->dsp);
+			disp->dsp = NULL;
+		}
+		return FALSE;
 	}
 
-	if (disp->servpid <= 0)
-		return;
-
 	gdm_debug ("gdm_server_reinit: Server for %s is about to be reinitialized!", disp->name);
 
-	/* This hack below is beautiful BTW */
+	if ( ! setup_server_wait (disp))
+		return FALSE;
 
-	gdm_sigchld_block_push ();
+	d->servstat = SERVER_PENDING;
 
-	pid = gdm_fork_extra ();
-	if (pid < 0) {
-		/* So much work just because we can't fork */
-		if (disp->servpid > 1)
-			kill (disp->servpid, SIGHUP);
+	if (disp->dsp != NULL) {
+		/* static because of the Setjmp */
+		static int (*old_xerror_handler)(Display *, XErrorEvent *) = NULL;
+		static int (*old_xioerror_handler)(Display *) = NULL;
 
-		gdm_sigchld_block_push ();
+		old_xerror_handler = NULL;
+		old_xioerror_handler = NULL;
 
-		/* HACK! the Xserver can't really tell us when it got the hup signal,
-		 * so we are really stuck just going to sleep for a bit hoping that
-		 * the kernel will tell the X server and that will run, else we will
-		 * get whacked ourselves after we open the connection and we'll think
-		 * it's an X screwup, which is really OK to happen and will just
-		 * restart the Xserver, it's just more nasty.  Oh how fun */
-		sleep (1);
-	} else if (pid == 0) {
-		Display *dsp;
+		/* Do note the interaction of this Setjmp and the signal
+	   	   handlers and the Setjmp in slave.c */
 
-		closelog ();
+		/* Long live Setjmp, DIE DIE DIE XSetIOErrorHandler */
 
-		signal (SIGCHLD, SIG_IGN);
-		signal (SIGTERM, SIG_DFL);
-		signal (SIGINT, SIG_DFL);
-		signal (SIGHUP, SIG_DFL);
+		if (Setjmp (reinitjmp) == 0)  {
+			/* come here and we'll whack the server and wait to get
+			   an xio error */
+			old_xerror_handler = XSetErrorHandler (ignore_xerror_handler);
+			old_xioerror_handler = XSetIOErrorHandler (jumpback_xioerror_handler);
 
-		/* close things */
-		gdm_close_all_descriptors (0 /* from */, -1 /* except */);
+			/* Now whack the server with a SIGHUP */
+			gdm_sigchld_block_push ();
+			if (disp->servpid > 1)
+				kill (disp->servpid, SIGHUP);
+			else
+				d->servstat = SERVER_DEAD;
+			gdm_sigchld_block_pop ();
 
-		gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
-		gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
-		gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
-
-		XSetErrorHandler (ignore_xerror_handler);
-		XSetIOErrorHandler (exit_xioerror_handler);
-	       
-		ve_setenv ("XAUTHORITY", d->authfile, TRUE);
-		dsp = XOpenDisplay (d->name);
-
-		/* Now whack the server with a SIGHUP */
-		if (disp->servpid > 1)
-			kill (disp->servpid, SIGHUP);
-
-		if (dsp == NULL) {
-			/* HACK, see note above */
-			sleep (1);
-			_exit (0);
-		} else {
-			/* Wait for an xioerror */
-			for (;;) {
-				XEvent event;
-				XNextEvent (dsp, &event);
+			/* the server is dead, weird */
+			if (disp->dsp != NULL) {
+				XCloseDisplay (disp->dsp);
+				disp->dsp = NULL;
 			}
 		}
-		/* anality */
-		_exit (0);
-	} else if (pid > 0) {
+		/* no more display */
+		disp->dsp = NULL;
+		XSetErrorHandler (old_xerror_handler);
+		XSetIOErrorHandler (old_xioerror_handler);
+	} else {
+		/* Now whack the server with a SIGHUP */
+		gdm_sigchld_block_push ();
+		if (disp->servpid > 1)
+			kill (disp->servpid, SIGHUP);
+		else
+			d->servstat = SERVER_DEAD;
 		gdm_sigchld_block_pop ();
+	}
 
-		gdm_wait_for_extra (NULL);
+	/* Wait for the SIGUSR1 */
+	do_server_wait (d);
+
+	if (d->servstat == SERVER_RUNNING) {
+#ifdef HAVE_FBCONSOLE
+		gdm_exec_fbconsole(d);
+#endif
+		return TRUE;
+        } else {
+		return FALSE;
 	}
 }
 
@@ -233,11 +263,16 @@ gdm_server_reinit (GdmDisplay *disp)
 void
 gdm_server_stop (GdmDisplay *disp)
 {
+    static gboolean waiting_for_server = FALSE;
+
     if (disp == NULL)
 	return;
 
     /* Kill our connection if one existed */
     if (disp->dsp != NULL) {
+	    /* on XDMCP servers first kill everything in sight */
+	    if (disp->type == TYPE_XDMCP)
+		    gdm_server_whack_clients (d);
 	    XCloseDisplay (disp->dsp);
 	    disp->dsp = NULL;
     }
@@ -257,13 +292,23 @@ gdm_server_stop (GdmDisplay *disp)
 
 	    /* avoid SIGCHLD race */
 	    gdm_sigchld_block_push ();
-
 	    servpid = disp->servpid;
+
+	    if (waiting_for_server) {
+		    gdm_error ("gdm_server_stop: Some problem killing server, whacking with SIGKILL");
+		    if (disp->servpid > 1)
+			    kill (disp->servpid, SIGKILL);
+
+	    } else {
+		    if (disp->servpid > 1 &&
+			kill (disp->servpid, SIGTERM) == 0) {
+			    waiting_for_server = TRUE;
+			    ve_waitpid_no_signal (disp->servpid, 0, 0);
+			    waiting_for_server = FALSE;
+		    }
+	    }
 	    disp->servpid = 0;
 
-	    if (servpid > 1 &&
-		kill (servpid, SIGTERM) == 0)
-		    ve_waitpid_no_signal (servpid, 0, 0);
 	    gdm_sigchld_block_pop ();
 
 	    gdm_server_whack_lockfile (disp);
@@ -429,36 +474,149 @@ display_vt (GdmDisplay *disp)
 	return -1;
 }
 
-static void
-check_child_status (void)
+static struct sigaction old_svr_wait_chld;
+static sigset_t old_svr_wait_mask;
+
+static gboolean
+setup_server_wait (GdmDisplay *d)
 {
-	int status;
-	pid_t pid;
+    struct sigaction usr1, chld;
+    sigset_t mask;
 
-	while ((pid = waitpid (-1, &status, WNOHANG)) > 0) {
-		gdm_debug ("check_child_status: %d died", pid);
+    if (pipe (server_signal_pipe) != 0) {
+	    gdm_error (_("%s: Error opening a pipe: %s"),
+		       "setup_server_wait", strerror (errno));
+	    return FALSE; 
+    }
+    server_signal_notified = FALSE;
 
-		if (WIFEXITED (status))
-			gdm_debug ("check_child_status: %d returned %d",
-				   (int)pid, (int)WEXITSTATUS (status));
-		if (WIFSIGNALED (status))
-			gdm_debug ("check_child_status: %d died of %d",
-				   (int)pid, (int)WTERMSIG (status));
+    /* Catch USR1 from X server */
+    usr1.sa_handler = gdm_server_usr1_handler;
+    usr1.sa_flags = SA_RESTART;
+    sigemptyset (&usr1.sa_mask);
 
-		if (pid == d->servpid) {
-			gdm_debug ("check_child_status: Got SIGCHLD from server, "
-				   "server abort");
+    if (sigaction (SIGUSR1, &usr1, NULL) < 0) {
+	    gdm_error (_("%s: Error setting up %s signal handler: %s"),
+		       "gdm_server_start", "USR1", strerror (errno));
+	    IGNORE_EINTR (close (server_signal_pipe[0]));
+	    IGNORE_EINTR (close (server_signal_pipe[1]));
+	    return FALSE;
+    }
 
-			d->servstat = SERVER_ABORT;	/* Server died unexpectedly */
-			d->servpid = 0;
+    /* Catch CHLD from X server */
+    chld.sa_handler = gdm_server_child_handler;
+    chld.sa_flags = SA_RESTART|SA_NOCLDSTOP;
+    sigemptyset (&chld.sa_mask);
 
-			server_signal_notified = TRUE;
-		} else if (pid == extra_process) {
-			/* an extra process died, yay! */
-			extra_process = 0;
-			extra_status = status;
-		}
-	}
+    if (sigaction (SIGCHLD, &chld, &old_svr_wait_chld) < 0) {
+	    gdm_error (_("%s: Error setting up %s signal handler: %s"),
+		       "gdm_server_start", "CHLD", strerror (errno));
+	    gdm_signal_ignore (SIGUSR1);
+	    IGNORE_EINTR (close (server_signal_pipe[0]));
+	    IGNORE_EINTR (close (server_signal_pipe[1]));
+	    return FALSE;
+    }
+
+    /* Set signal mask */
+    sigemptyset (&mask);
+    sigaddset (&mask, SIGUSR1);
+    sigaddset (&mask, SIGCHLD);
+    sigprocmask (SIG_UNBLOCK, &mask, &old_svr_wait_mask);
+
+    return TRUE;
+}
+
+static void
+do_server_wait (GdmDisplay *d)
+{
+    /* Wait for X server to send ready signal */
+    if (d->servstat == SERVER_PENDING) {
+	    if (d->server_uid != 0 && ! d->handled) {
+		    /* FIXME: If not handled, we just don't know, so
+		     * just wait a few seconds and hope things just work,
+		     * fortunately there is no such case yet and probably
+		     * never will, but just for code anality's sake */
+		    gdm_sleep_no_signal (5);
+	    } else if (d->server_uid != 0) {
+		    int i;
+
+		    /* FIXME: This is not likely to work in reinit,
+		       but we never reinit Xnest servers nowdays,
+		       so that's fine */
+
+		    /* if we're running the server as a non-root, we can't
+		     * use USR1 of course, so try openning the display 
+		     * as a test, but the */
+
+		    /* just in case it's set */
+		    ve_unsetenv ("XAUTHORITY");
+
+		    gdm_auth_set_local_auth (d);
+
+		    for (i = 0;
+			 d->dsp == NULL &&
+			 d->servstat == SERVER_PENDING &&
+			 i < SERVER_WAIT_ALARM;
+			 i++) {
+			    d->dsp = XOpenDisplay (d->name);
+			    if (d->dsp == NULL)
+				    gdm_sleep_no_signal (1);
+			    else
+				    d->servstat = SERVER_RUNNING;
+		    }
+		    if (d->dsp == NULL &&
+			/* Note: we could have still gotten a SIGCHLD */
+			d->servstat == SERVER_PENDING) {
+			    d->servstat = SERVER_TIMEOUT;
+		    }
+	    } else {
+		    time_t t = time (NULL);
+
+		    gdm_debug ("do_server_wait: Before mainloop waiting for server");
+
+		    do {
+			    fd_set rfds;
+			    struct timeval tv;
+
+			    /* Wait up to SERVER_WAIT_ALARM seconds. */
+			    tv.tv_sec = MAX (1, SERVER_WAIT_ALARM - (time(NULL) - t));
+			    tv.tv_usec = 0;
+
+			    FD_ZERO (&rfds);
+			    FD_SET (server_signal_pipe[0], &rfds);
+
+			    if (select (server_signal_pipe[0]+1, &rfds, NULL, NULL, &tv) > 0) {
+				    char buf[4];
+				    /* read the Yay! */
+				    IGNORE_EINTR (read (server_signal_pipe[0], buf, 4));
+			    }
+			    if ( ! server_signal_notified &&
+				t + SERVER_WAIT_ALARM < time (NULL)) {
+				    gdm_debug ("do_server_wait: Server timeout");
+				    d->servstat = SERVER_TIMEOUT;
+				    server_signal_notified = TRUE;
+			    }
+			    if (d->servpid <= 1) {
+				    d->servstat = SERVER_ABORT;
+				    server_signal_notified = TRUE;
+			    }
+		    } while ( ! server_signal_notified);
+
+		    gdm_debug ("gdm_server_start: After mainloop waiting for server");
+	    }
+    }
+
+    /* restore default handlers */
+    gdm_signal_ignore (SIGUSR1);
+    sigaction (SIGCHLD, &old_svr_wait_chld, NULL);
+    sigprocmask (SIG_SETMASK, &old_svr_wait_mask, NULL);
+
+    IGNORE_EINTR (close (server_signal_pipe[0]));
+    IGNORE_EINTR (close (server_signal_pipe[1]));
+
+    if (d->servpid <= 1) {
+	    d->servstat = SERVER_ABORT;
+    }
 }
 
 /**
@@ -472,9 +630,6 @@ gboolean
 gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
 		  int min_flexi_disp, int flexi_retries)
 {
-    struct sigaction usr1, chld, alrm;
-    struct sigaction old_usr1, old_chld, old_alrm;
-    sigset_t mask, oldmask;
     int flexi_disp = 20;
     char *vtarg = NULL;
     int vtfd = -1, vt = -1;
@@ -516,71 +671,8 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
     gdm_slave_send_string (GDM_SOP_COOKIE, d->cookie);
     ve_setenv ("DISPLAY", d->name, TRUE);
 
-    if (pipe (server_signal_pipe) != 0) {
-	    gdm_error (_("%s: Error opening a pipe: %s"),
-		       "gdm_server_start", strerror (errno));
-	    return FALSE; 
-    }
-    server_signal_notified = FALSE;
-
-    /* Catch USR1 from X server */
-    usr1.sa_handler = gdm_server_usr1_handler;
-    usr1.sa_flags = SA_RESTART|SA_RESETHAND;
-    sigemptyset (&usr1.sa_mask);
-
-    if (sigaction (SIGUSR1, &usr1, &old_usr1) < 0) {
-	    gdm_error (_("%s: Error setting up USR1 signal handler: %s"),
-		       "gdm_server_start", strerror (errno));
-	    close (server_signal_pipe[0]);
-	    close (server_signal_pipe[1]);
+    if ( ! setup_server_wait (d))
 	    return FALSE;
-    }
-
-    /* Catch CHLD from X server */
-    chld.sa_handler = gdm_server_child_handler;
-    chld.sa_flags = SA_RESTART|SA_RESETHAND;
-    sigemptyset (&chld.sa_mask);
-
-    if (sigaction (SIGCHLD, &chld, &old_chld) < 0) {
-	    gdm_error (_("%s: Error setting up CHLD signal handler: %s"),
-		       "gdm_server_start", strerror (errno));
-	    sigaction (SIGUSR1, &old_usr1, NULL);
-	    close (server_signal_pipe[0]);
-	    close (server_signal_pipe[1]);
-	    return FALSE;
-    }
-
-    /* Catch ALRM from X server */
-    alrm.sa_handler = gdm_server_alarm_handler;
-    alrm.sa_flags = SA_RESTART|SA_RESETHAND;
-    sigemptyset (&alrm.sa_mask);
-
-    if (sigaction (SIGALRM, &alrm, &old_alrm) < 0) {
-	    gdm_error (_("%s: Error setting up ALRM signal handler: %s"),
-		       "gdm_server_start", strerror (errno));
-	    sigaction (SIGUSR1, &old_usr1, NULL);
-	    sigaction (SIGCHLD, &old_chld, NULL);
-	    close (server_signal_pipe[0]);
-	    close (server_signal_pipe[1]);
-	    return FALSE;
-    }
-
-    /* Set signal mask */
-    sigemptyset (&mask);
-    sigaddset (&mask, SIGUSR1);
-    sigaddset (&mask, SIGCHLD);
-    sigaddset (&mask, SIGALRM);
-    sigprocmask (SIG_UNBLOCK, &mask, &oldmask);
-
-    /* We add a timeout in case the X server fails to start. This
-     * might happen because X servers take a while to die, close their
-     * sockets etc. If the old X server isn't completely dead, the new
-     * one will fail and we'll hang here forever */
-
-    /* Only do alarm if server will be run as root */
-    if (d->server_uid == 0) {
-	    alarm (SERVER_WAIT_ALARM);
-    }
 
     d->servstat = SERVER_DEAD;
 
@@ -594,80 +686,13 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
 
     /* we can now use d->handled since that's set up above */
 
-    /* Wait for X server to send ready signal */
-    if (d->servstat == SERVER_STARTED) {
-	    if (d->server_uid != 0 && ! d->handled) {
-		    alarm (0);
-		    /* FIXME: If not handled, we just don't know, so
-		     * just wait a few seconds and hope things just work,
-		     * fortunately there is no such case yet and probably
-		     * never will, but just for code anality's sake */
-		    sleep (5);
-		    /* In case we got a SIGCHLD */
-		    check_child_status ();
-	    } else if (d->server_uid != 0) {
-		    int i;
-
-		    alarm (0);
-
-		    /* if we're running the server as a non-root, we can't
-		     * use USR1 of course, so try openning the display 
-		     * as a test, but the */
-
-		    /* In case we got a SIGCHLD */
-		    check_child_status ();
-
-		    ve_setenv ("XAUTHORITY", d->authfile, TRUE);
-		    for (i = 0;
-			 d->dsp == NULL &&
-			 d->servstat == SERVER_STARTED &&
-			 i < SERVER_WAIT_ALARM;
-			 i++) {
-			    d->dsp = XOpenDisplay (d->name);
-			    if (d->dsp == NULL)
-				    sleep (1);
-			    else
-				    d->servstat = SERVER_RUNNING;
-
-			    /* In case we got a SIGCHLD */
-			    check_child_status ();
-		    }
-		    ve_unsetenv ("XAUTHORITY");
-		    if (d->dsp == NULL &&
-			/* Note: we could have still gotten a SIGCHLD */
-			d->servstat == SERVER_STARTED) {
-			    d->servstat = SERVER_TIMEOUT;
-		    }
-	    } else {
-		    fd_set rfds;
-
-		    gdm_debug ("gdm_server_start: Before mainloop waiting for server");
-
-		    FD_ZERO (&rfds);
-		    FD_SET (server_signal_pipe[0], &rfds);
-
-		    do {
-			    select (server_signal_pipe[0]+1, &rfds, NULL, NULL, NULL);
-			    /* In case we got a SIGCHLD */
-			    check_child_status ();
-		    } while ( ! server_signal_notified);
-
-		    gdm_debug ("gdm_server_start: After mainloop waiting for server");
-	    }
-    }
-
-    /* In case we got a SIGCHLD */
-    check_child_status ();
-
-    /* Unset alarm */
-    if (d->server_uid == 0) {
-	    alarm (0);
-    }
+    do_server_wait (d);
 
     /* If we were holding a vt open for the server, close it now as it has
      * already taken the bait. */
-    if (vtfd > 0)
-	    close (vtfd);
+    if (vtfd > 0) {
+	    IGNORE_EINTR (close (vtfd));
+    }
 
     switch (d->servstat) {
 
@@ -682,16 +707,6 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
     case SERVER_RUNNING:
 	    gdm_debug ("gdm_server_start: Completed %s!", d->name);
 
-	    sigprocmask (SIG_SETMASK, &oldmask, NULL);
-
-	    /* restore default handlers */
-	    sigaction (SIGUSR1, &old_usr1, NULL);
-	    sigaction (SIGCHLD, &old_chld, NULL);
-	    sigaction (SIGALRM, &old_alrm, NULL);
-
-	    close (server_signal_pipe[0]);
-	    close (server_signal_pipe[1]);
-
 	    if (SERVER_IS_FLEXI (d))
 		    gdm_slave_send_num (GDM_SOP_FLEXI_OK, 0 /* bogus */);
 	    if (d->type == TYPE_LOCAL ||
@@ -705,6 +720,10 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
 			    gdm_slave_send_num (GDM_SOP_VT_NUM, d->vt);
 	    }
 
+#ifdef HAVE_FBCONSOLE
+            gdm_exec_fbconsole(d);
+#endif
+
 	    return TRUE;
     default:
 	    break;
@@ -730,19 +749,6 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
     /* We will rebake cookies anyway, so wipe these */
     gdm_server_wipe_cookies (disp);
 
-    sigprocmask (SIG_SETMASK, &oldmask, NULL);
-
-    /* In case we got a SIGCHLD */
-    check_child_status ();
-
-    /* restore default handlers */
-    sigaction (SIGUSR1, &old_usr1, NULL);
-    sigaction (SIGCHLD, &old_chld, NULL);
-    sigaction (SIGALRM, &old_alrm, NULL);
-
-    close (server_signal_pipe[0]);
-    close (server_signal_pipe[1]);
-
     if (disp->type == TYPE_FLEXI_XNEST &&
 	display_xnest_no_connect (disp)) {
 	    gdm_slave_send_num (GDM_SOP_FLEXI_ERR,
@@ -810,6 +816,27 @@ gdm_server_checklog (GdmDisplay *disp)
 	}
 }
 
+/* somewhat safer rename (safer if the log dir is unsafe), may in fact
+   lose the file though, it guarantees that a is gone, but not that
+   b exists */
+static void
+safer_rename (const char *a, const char *b)
+{
+	errno = 0;
+	if (link (a, b) < 0) {
+		if (errno == EEXIST) {
+			IGNORE_EINTR (unlink (a));
+			return;
+		} 
+		IGNORE_EINTR (unlink (b));
+		/* likely this system doesn't support hard links */
+		rename (a, b);
+		IGNORE_EINTR (unlink (a));
+		return;
+	}
+	IGNORE_EINTR (unlink (a));
+}
+
 static void
 rotate_logs (const char *dname)
 {
@@ -821,11 +848,11 @@ rotate_logs (const char *dname)
 	char *fname = g_strconcat (GdmLogDir, "/", dname, ".log", NULL);
 
 	/* Rotate the logs (keep 4 last) */
-	unlink (fname4);
-	rename (fname3, fname4);
-	rename (fname2, fname3);
-	rename (fname1, fname2);
-	rename (fname, fname1);
+	IGNORE_EINTR (unlink (fname4));
+	safer_rename (fname3, fname4);
+	safer_rename (fname2, fname3);
+	safer_rename (fname1, fname2);
+	safer_rename (fname, fname1);
 
 	g_free (fname4);
 	g_free (fname3);
@@ -955,9 +982,10 @@ gdm_server_resolve_command_line (GdmDisplay *disp,
 static void 
 gdm_server_spawn (GdmDisplay *d, const char *vtarg)
 {
-    struct sigaction ign_signal, dfl_signal;
+    struct sigaction ign_signal;
     sigset_t mask;
     gchar **argv = NULL;
+    char *logfile;
     int logfd;
     char *command;
     pid_t pid;
@@ -967,7 +995,7 @@ gdm_server_spawn (GdmDisplay *d, const char *vtarg)
 	    return;
     }
 
-    d->servstat = SERVER_STARTED;
+    d->servstat = SERVER_PENDING;
 
     gdm_sigchld_block_push ();
 
@@ -1000,16 +1028,13 @@ gdm_server_spawn (GdmDisplay *d, const char *vtarg)
     switch (pid) {
 	
     case 0:
-        alarm (0);
-
-	/* Close the XDMCP fd inherited by the daemon process */
-	if (GdmXdmcp)
-		gdm_xdmcp_close();
+	/* the pops whacked mask again */
+        gdm_unset_signals ();
 
 	closelog ();
 
 	/* close things */
-	gdm_close_all_descriptors (0 /* from */, -1 /* except */);
+	gdm_close_all_descriptors (0 /* from */, -1 /* except */, -1 /* except2 */);
 
 	/* No error checking here - if it's messed the best response
          * is to ignore & try to continue */
@@ -1023,12 +1048,13 @@ gdm_server_spawn (GdmDisplay *d, const char *vtarg)
 	rotate_logs (d->name);
 
         /* Log all output from spawned programs to a file */
-	logfd = open (g_strconcat (GdmLogDir, "/", d->name, ".log", NULL),
-		      O_CREAT|O_TRUNC|O_WRONLY, 0644);
+	logfile = g_strconcat (GdmLogDir, "/", d->name, ".log", NULL);
+	IGNORE_EINTR (unlink (logfile));
+	logfd = open (logfile, O_CREAT|O_TRUNC|O_WRONLY|O_EXCL, 0644);
 
 	if (logfd != -1) {
-		dup2 (logfd, 1);
-		dup2 (logfd, 2);
+		IGNORE_EINTR (dup2 (logfd, 1));
+		IGNORE_EINTR (dup2 (logfd, 2));
         } else {
 		gdm_error (_("%s: Could not open logfile for display %s!"),
 			   "gdm_server_spawn", d->name);
@@ -1040,10 +1066,13 @@ gdm_server_spawn (GdmDisplay *d, const char *vtarg)
 	ign_signal.sa_flags = SA_RESTART;
 	sigemptyset (&ign_signal.sa_mask);
 
-	if (sigaction (SIGUSR1, &ign_signal, NULL) < 0) {
-		gdm_error (_("%s: Error setting %s to %s"),
-			   "gdm_server_spawn", "USR1", "SIG_IGN");
-		_exit (SERVER_ABORT);
+	if (d->server_uid == 0) {
+		/* only set this if we can actually listen */
+		if (sigaction (SIGUSR1, &ign_signal, NULL) < 0) {
+			gdm_error (_("%s: Error setting %s to %s"),
+				   "gdm_server_spawn", "USR1", "SIG_IGN");
+			_exit (SERVER_ABORT);
+		}
 	}
 	if (sigaction (SIGTTIN, &ign_signal, NULL) < 0) {
 		gdm_error (_("%s: Error setting %s to %s"),
@@ -1056,21 +1085,8 @@ gdm_server_spawn (GdmDisplay *d, const char *vtarg)
 		_exit (SERVER_ABORT);
 	}
 
-	/* And HUP and TERM should be at default */
-	dfl_signal.sa_handler = SIG_DFL;
-	dfl_signal.sa_flags = SA_RESTART;
-	sigemptyset (&dfl_signal.sa_mask);
-
-	if (sigaction (SIGHUP, &dfl_signal, NULL) < 0) {
-		gdm_error (_("%s: Error setting %s to %s"),
-			   "gdm_server_spawn", "HUP", "SIG_DFL");
-		_exit (SERVER_ABORT);
-	}
-	if (sigaction (SIGTERM, &dfl_signal, NULL) < 0) {
-		gdm_error (_("%s: Error setting %s to %s"),
-			   "gdm_server_spawn", "TERM", "SIG_DFL");
-		_exit (SERVER_ABORT);
-	}
+	/* And HUP and TERM are at SIG_DFL from gdm_unset_signals,
+	   we also have an empty mask and all that fun stuff */
 
 	/* unblock signals (especially HUP/TERM/USR1) so that we
 	 * can control the X server */
@@ -1158,7 +1174,7 @@ gdm_server_spawn (GdmDisplay *d, const char *vtarg)
 		setgroups (1, groups);
 	}
 
-	execv (argv[0], argv);
+	IGNORE_EINTR (execv (argv[0], argv));
 	
 	gdm_error (_("%s: Xserver not found: %s"), 
 		   "gdm_server_spawn", command);
@@ -1203,31 +1219,7 @@ gdm_server_usr1_handler (gint sig)
 
     server_signal_notified = TRUE;
     /* this will quit the select */
-    write (server_signal_pipe[1], "Yay!", 4);
-
-    gdm_in_signal--;
-}
-
-
-/**
- * gdm_server_alarm_handler:
- * @sig: Signal value
- *
- * Server start timeout handler
- */
-
-static void 
-gdm_server_alarm_handler (gint signal)
-{
-    gdm_in_signal++;
-
-    d->servstat = SERVER_TIMEOUT; /* Server didn't start */
-
-    gdm_debug ("gdm_server_alarm_handler: Got SIGALRM, server abort");
-
-    server_signal_notified = TRUE;
-    /* this will quit the select */
-    write (server_signal_pipe[1], "Yay!", 4);
+    IGNORE_EINTR (write (server_signal_pipe[1], "Yay!", 4));
 
     gdm_in_signal--;
 }
@@ -1247,8 +1239,11 @@ gdm_server_child_handler (int signal)
 
 	gdm_debug ("gdm_server_child_handler: Got SIGCHLD");
 
+	/* go to the main child handler */
+	gdm_slave_child_handler (signal);
+
 	/* this will quit the select */
-	write (server_signal_pipe[1], "Yay!", 4);
+	IGNORE_EINTR (write (server_signal_pipe[1], "Yay!", 4));
 
 	gdm_in_signal--;
 }
@@ -1316,6 +1311,10 @@ gdm_server_alloc (gint id, const gchar *command)
 
     d->slave_notify_fd = -1;
     d->master_notify_fd = -1;
+
+    d->xsession_errors_bytes = 0;
+    d->xsession_errors_fd = -1;
+    d->session_output_fd = -1;
     
     return d;
 }
