@@ -96,6 +96,7 @@
 #include "misc.h"
 #include "choose.h"
 #include "xdmcp.h"
+#include "cookie.h"
 
 int gdm_xdmcpfd = -1;
 
@@ -132,6 +133,8 @@ extern gchar *GdmWilling;	/* The willing script */
 
 extern gboolean GdmXdmcp;	/* xdmcp enabled */
 
+extern gboolean GdmDebug;	/* debug enabled */
+
 /* Local prototypes */
 static gboolean gdm_xdmcp_decode_packet (GIOChannel *source,
 					 GIOCondition cond,
@@ -159,7 +162,7 @@ static void gdm_xdmcp_send_accept (GdmHostent *he /* eaten and freed */,
 static void gdm_xdmcp_send_decline (struct sockaddr_in *clnt_sa, const char *reason);
 static void gdm_xdmcp_send_refuse (struct sockaddr_in *clnt_sa, CARD32 sessid);
 static void gdm_xdmcp_send_failed (struct sockaddr_in *clnt_sa, CARD32 sessid);
-static void gdm_xdmcp_send_alive (struct sockaddr_in *clnt_sa, CARD32 sessid);
+static void gdm_xdmcp_send_alive (struct sockaddr_in *clnt_sa, CARD16 dspnum, CARD32 sessid);
 static void gdm_xdmcp_send_managed_forward (struct sockaddr_in *clnt_sa,
 					    struct sockaddr_in *origin);
 static void gdm_xdmcp_send_got_managed_forward (struct sockaddr_in *clnt_sa,
@@ -169,6 +172,8 @@ static GdmDisplay *gdm_xdmcp_display_alloc (struct in_addr *addr,
 					    GdmHostent *he /* eaten and freed */,
 					    int displaynum);
 static GdmDisplay *gdm_xdmcp_display_lookup (CARD32 sessid);
+static GdmDisplay *gdm_xdmcp_display_lookup_by_host (struct in_addr *addr,
+						     int dspnum);
 static void gdm_xdmcp_display_dispose_check (const gchar *name);
 static void gdm_xdmcp_displays_check (void);
 static int gdm_xdmcp_displays_from_host (struct in_addr *addr);
@@ -188,7 +193,7 @@ typedef struct {
 } ManagedForward;
 #define MANAGED_FORWARD_INTERVAL 1500 /* 1.5 seconds */
 
-static GList *managed_forwards = NULL;
+static GSList *managed_forwards = NULL;
 
 
 /* 
@@ -229,6 +234,23 @@ gdm_xdmcp_displays_from_host (struct in_addr *addr)
 	return count;
 }
 
+static GdmDisplay *
+gdm_xdmcp_display_lookup_by_host (struct in_addr *addr, int dspnum)
+{
+	GSList *li;
+
+	for (li = displays; li != NULL; li = li->next) {
+		GdmDisplay *disp = li->data;
+		if (disp->type == TYPE_XDMCP &&
+		    memcmp (&disp->addr, addr, sizeof (struct in_addr)) == 0 &&
+		    disp->dispnum == dspnum) {
+			return disp;
+		}
+	}
+
+	return NULL;
+}
+
 
 gboolean
 gdm_xdmcp_init (void)
@@ -239,12 +261,12 @@ gdm_xdmcp_init (void)
 
     if ( ! GdmXdmcp)
 	    return TRUE;
-    
-    globsessid = time (NULL);
+
+    globsessid = g_random_int ();
     
     /* Fetch and store local hostname in XDMCP friendly format */
     hostbuf[1023] = '\0';
-    if (gethostname (hostbuf, 1023) != 0) {
+    if G_UNLIKELY (gethostname (hostbuf, 1023) != 0) {
 	gdm_error (_("%s: Could not get server hostname: %s!"), 
 		   "gdm_xdmcp_init", strerror (errno));
 	strcmp (hostbuf, "localhost.localdomain");
@@ -265,7 +287,7 @@ gdm_xdmcp_init (void)
     /* Open socket for communications */
     gdm_xdmcpfd = socket (AF_INET, SOCK_DGRAM, 0); /* UDP */
     
-    if (gdm_xdmcpfd < 0) {
+    if G_UNLIKELY (gdm_xdmcpfd < 0) {
 	gdm_error (_("%s: Could not create socket!"), "gdm_xdmcp_init");
 	GdmXdmcp = FALSE;
 	return FALSE;
@@ -275,7 +297,7 @@ gdm_xdmcp_init (void)
     serv_sa.sin_port = htons (GdmPort); /* UDP 177 */
     serv_sa.sin_addr.s_addr = htonl (INADDR_ANY);
     
-    if (bind (gdm_xdmcpfd, (struct sockaddr*) &serv_sa, sizeof (serv_sa)) == -1) {
+    if G_UNLIKELY (bind (gdm_xdmcpfd, (struct sockaddr*) &serv_sa, sizeof (serv_sa)) == -1) {
 	gdm_error (_("%s: Could not bind to XDMCP socket!"), "gdm_xdmcp_init");
 	gdm_xdmcp_close ();
 	GdmXdmcp = FALSE;
@@ -333,32 +355,41 @@ gdm_xdmcp_decode_packet (GIOChannel *source, GIOCondition cond, gpointer data)
     static const char * const gdm_opcode_names[] = {
 	"MANAGED_FORWARD", "GOT_MANAGED_FORWARD"
     };
-    
-    if (!XdmcpFill (gdm_xdmcpfd, &buf, (XdmcpNetaddr)&clnt_sa, &sa_len)) {
+
+    /* packets come at somewhat random times */
+    gdm_random_tick ();
+
+    if (cond != G_IO_IN)
+	    gdm_debug ("gdm_xdmcp_decode_packet: GIOCondition %d", (int)cond);
+
+    if ( ! (cond & G_IO_IN)) 
+        return TRUE;
+
+    if G_UNLIKELY (!XdmcpFill (gdm_xdmcpfd, &buf, (XdmcpNetaddr)&clnt_sa, &sa_len)) {
 	gdm_error (_("%s: Could not create XDMCP buffer!"),
 		   "gdm_xdmcp_decode_packet");
 	return TRUE;
     }
     
-    if (!XdmcpReadHeader (&buf, &header)) {
+    if G_UNLIKELY (!XdmcpReadHeader (&buf, &header)) {
 	gdm_error (_("%s: Could not read XDMCP header!"),
 		   "gdm_xdmcp_decode_packet");
 	return TRUE;
     }
     
-    if (header.version != XDM_PROTOCOL_VERSION &&
-	header.version != GDM_XDMCP_PROTOCOL_VERSION) {
+    if G_UNLIKELY (header.version != XDM_PROTOCOL_VERSION &&
+		   header.version != GDM_XDMCP_PROTOCOL_VERSION) {
 	gdm_error (_("%s: Incorrect XDMCP version!"),
 		   "gdm_xdmcp_decode_packet");
 	return TRUE;
     }
 
     if (header.opcode <= ALIVE)
-	    gdm_debug ("gdm_xdmcp_decode: Received opcode %s from client %s", 
+	    gdm_debug ("gdm_xdmcp_decode_packet: Received opcode %s from client %s", 
 		       opcode_names[header.opcode], inet_ntoa (clnt_sa.sin_addr));
     if (header.opcode >= GDM_XDMCP_FIRST_OPCODE &&
         header.opcode < GDM_XDMCP_LAST_OPCODE)
-	    gdm_debug ("gdm_xdmcp_decode: Received opcode %s from client %s", 
+	    gdm_debug ("gdm_xdmcp_decode_packet: Received opcode %s from client %s", 
 		       gdm_opcode_names[header.opcode - GDM_XDMCP_FIRST_OPCODE],
 		       inet_ntoa (clnt_sa.sin_addr));
 
@@ -381,7 +412,6 @@ gdm_xdmcp_decode_packet (GIOChannel *source, GIOCondition cond, gpointer data)
 	break;
 	
     case REQUEST:
-	gdm_xdmcp_displays_check(); /* Purge pending displays */
 	gdm_xdmcp_handle_request (&clnt_sa, header.length);
 	break;
 	
@@ -421,7 +451,7 @@ gdm_xdmcp_handle_query (struct sockaddr_in *clnt_sa, gint len, gint type)
 	       type, inet_ntoa (clnt_sa->sin_addr));
     
     /* Extract array of authentication names from Xdmcp packet */
-    if (! XdmcpReadARRAYofARRAY8 (&buf, &clnt_authlist)) {
+    if G_UNLIKELY (! XdmcpReadARRAYofARRAY8 (&buf, &clnt_authlist)) {
 	gdm_error (_("%s: Could not extract authlist from packet"),
 		   "gdm_xdmcp_handle_query"); 
 	return;
@@ -429,12 +459,15 @@ gdm_xdmcp_handle_query (struct sockaddr_in *clnt_sa, gint len, gint type)
     
     /* Crude checksumming */
     for (i = 0 ; i < clnt_authlist.length ; i++) {
-	gdm_debug ("gdm_xdmcp_handle_query: authlist: %s",
-		   (char *)clnt_authlist.data);
-	explen += 2+clnt_authlist.data[i].length;
+	    if G_UNLIKELY (GdmDebug) {
+		    char *s = g_strndup (clnt_authlist.data[i].data, clnt_authlist.length);
+		    gdm_debug ("gdm_xdmcp_handle_query: authlist: %s", ve_sure_string (s));
+		    g_free (s);
+	    }
+	    explen += 2+clnt_authlist.data[i].length;
     }
     
-    if (len != explen) {
+    if G_UNLIKELY (len != explen) {
 	gdm_error (_("%s: Error in checksum"),
 		   "gdm_xdmcp_handle_query"); 
 	XdmcpDisposeARRAYofARRAY8 (&clnt_authlist);
@@ -519,8 +552,8 @@ gdm_xdmcp_send_forward_query (GdmIndirectDisplay *id,
 	ARRAY8 address;
 	ARRAY8 port;
 
-	g_assert (id != NULL);
-	g_assert (id->chosen_host != NULL);
+	gdm_assert (id != NULL);
+	gdm_assert (id->chosen_host != NULL);
 
 	gdm_debug ("gdm_xdmcp_send_forward_query: Sending forward query to %s",
 		   inet_ntoa (*id->chosen_host));
@@ -632,8 +665,9 @@ gdm_forward_query_lookup (struct sockaddr_in *clnt_sa)
 		if (q->acctime > 0 &&
 		    curtime > q->acctime + GDM_FORWARD_QUERY_TIMEOUT)	{
 			gdm_debug ("gdm_forward_query_lookup: Disposing stale forward query from %s",
-				   inet_ntoa (clnt_sa->sin_addr));
+				   inet_ntoa (q->dsp_sa->sin_addr));
 			gdm_forward_query_dispose (q);
+			continue;
 		}
 
 	}
@@ -676,16 +710,24 @@ gdm_xdmcp_handle_forward_query (struct sockaddr_in *clnt_sa, gint len)
     ARRAYofARRAY8 clnt_authlist;
     gint i = 0, explen = 1;
     struct sockaddr_in disp_sa = {0};
+
+    /* Check with tcp_wrappers if client is allowed to access */
+    if (! gdm_xdmcp_host_allow (clnt_sa)) {
+	    gdm_error ("%s: Got FORWARD_QUERY from banned host %s", 
+		       "gdm_xdmcp_handle_forward query",
+		       inet_ntoa (clnt_sa->sin_addr));
+	    return;
+    }
     
     /* Read display address */
-    if (! XdmcpReadARRAY8 (&buf, &clnt_addr)) {
+    if G_UNLIKELY (! XdmcpReadARRAY8 (&buf, &clnt_addr)) {
 	gdm_error (_("%s: Could not read display address"),
 		   "gdm_xdmcp_handle_forward_query");
 	return;
     }
     
     /* Read display port */
-    if (! XdmcpReadARRAY8 (&buf, &clnt_port)) {
+    if G_UNLIKELY (! XdmcpReadARRAY8 (&buf, &clnt_port)) {
 	XdmcpDisposeARRAY8 (&clnt_addr);
 	gdm_error (_("%s: Could not read display port number"),
 		   "gdm_xdmcp_handle_forward_query");
@@ -693,7 +735,7 @@ gdm_xdmcp_handle_forward_query (struct sockaddr_in *clnt_sa, gint len)
     }
     
     /* Extract array of authentication names from Xdmcp packet */
-    if (! XdmcpReadARRAYofARRAY8 (&buf, &clnt_authlist)) {
+    if G_UNLIKELY (! XdmcpReadARRAYofARRAY8 (&buf, &clnt_authlist)) {
 	XdmcpDisposeARRAY8 (&clnt_addr);
 	XdmcpDisposeARRAY8 (&clnt_port);
 	gdm_error (_("%s: Could not extract authlist from packet"),
@@ -707,25 +749,28 @@ gdm_xdmcp_handle_forward_query (struct sockaddr_in *clnt_sa, gint len)
     explen += 2+clnt_port.length;
     
     for (i = 0 ; i < clnt_authlist.length ; i++) {
-	gdm_debug ("gdm_xdmcp_handle_forward_query: authlist: %s",
-		   (char *)clnt_authlist.data);
-	explen += 2+clnt_authlist.data[i].length;
+	    if G_UNLIKELY (GdmDebug) {
+		    char *s = g_strndup (clnt_authlist.data[i].data, clnt_authlist.length);
+		    gdm_debug ("gdm_xdmcp_handle_forward_query: authlist: %s", ve_sure_string (s));
+		    g_free (s);
+	    }
+	    explen += 2+clnt_authlist.data[i].length;
     }
     
-    if (len != explen) {
+    if G_UNLIKELY (len != explen) {
 	gdm_error (_("%s: Error in checksum"),
 		   "gdm_xdmcp_handle_forward_query"); 
 	goto out;
     }
     
-    if (clnt_port.length != 2 ||
-	clnt_addr.length != 4) {
+    if G_UNLIKELY (clnt_port.length != 2 ||
+		   clnt_addr.length != 4) {
 	    gdm_error (_("%s: Bad address"),
 		       "gdm_xdmcp_handle_forward_query"); 
 	    goto out;
     }
 
-    g_assert (4 == sizeof (struct in_addr));
+    gdm_assert (4 == sizeof (struct in_addr));
     gdm_xdmcp_whack_queued_managed_forwards
 	    (clnt_sa, (struct in_addr *)clnt_addr.data);
 
@@ -826,6 +871,12 @@ gdm_xdmcp_send_unwilling (struct sockaddr_in *clnt_sa, gint type)
 {
     ARRAY8 status;
     XdmcpHeader header;
+    static time_t last_time = 0;
+
+    /* only send at most one packet per second,
+       no harm done if we don't send it at all */
+    if (last_time + 1 >= time (NULL))
+	    return;
     
     gdm_debug ("gdm_xdmcp_send_unwilling: Sending UNWILLING to %s", inet_ntoa (clnt_sa->sin_addr));
     
@@ -845,6 +896,8 @@ gdm_xdmcp_send_unwilling (struct sockaddr_in *clnt_sa, gint type)
     XdmcpWriteARRAY8 (&buf, &status);
     XdmcpFlush (gdm_xdmcpfd, &buf, (XdmcpNetaddr)clnt_sa,
 		(int)sizeof (struct sockaddr_in));
+
+    last_time = time (NULL);
 }
 
 static void
@@ -882,7 +935,7 @@ managed_forward_handler (gpointer data)
 						       &(mf->origin));
 	mf->times ++;
 	if (gdm_xdmcpfd <= 0 || mf->times >= 2) {
-		managed_forwards = g_list_remove (managed_forwards, mf);
+		managed_forwards = g_slist_remove (managed_forwards, mf);
 		mf->handler = 0;
 		/* mf freed by glib */
 		return FALSE;
@@ -908,7 +961,7 @@ gdm_xdmcp_send_managed_forward (struct sockaddr_in *clnt_sa,
 					  managed_forward_handler,
 					  mf,
 					  (GDestroyNotify) g_free);
-	managed_forwards = g_list_prepend (managed_forwards, mf);
+	managed_forwards = g_slist_prepend (managed_forwards, mf);
 }
 
 static void
@@ -961,23 +1014,25 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
 		   inet_ntoa (clnt_sa->sin_addr));
 	return;
     }
+
+    gdm_xdmcp_displays_check(); /* Purge pending displays */
     
     /* Remote display number */
-    if (! XdmcpReadCARD16 (&buf, &clnt_dspnum)) {
+    if G_UNLIKELY (! XdmcpReadCARD16 (&buf, &clnt_dspnum)) {
 	gdm_error (_("%s: Could not read Display Number"),
 		   "gdm_xdmcp_handle_request");
 	return;
     }
     
     /* We don't care about connection type. Address says it all */
-    if (! XdmcpReadARRAY16 (&buf, &clnt_conntyp)) {
+    if G_UNLIKELY (! XdmcpReadARRAY16 (&buf, &clnt_conntyp)) {
 	gdm_error (_("%s: Could not read Connection Type"),
 		   "gdm_xdmcp_handle_request");
 	return;
     }
     
     /* This is TCP/IP - we don't care */
-    if (! XdmcpReadARRAYofARRAY8 (&buf, &clnt_addr)) {
+    if G_UNLIKELY (! XdmcpReadARRAYofARRAY8 (&buf, &clnt_addr)) {
 	gdm_error (_("%s: Could not read Client Address"),
 		   "gdm_xdmcp_handle_request");
         XdmcpDisposeARRAY16 (&clnt_conntyp);
@@ -985,7 +1040,7 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
     }
     
     /* Read authentication type */
-    if (! XdmcpReadARRAY8 (&buf, &clnt_authname)) {
+    if G_UNLIKELY (! XdmcpReadARRAY8 (&buf, &clnt_authname)) {
 	gdm_error (_("%s: Could not read Authentication Names"),
 		   "gdm_xdmcp_handle_request");
 	XdmcpDisposeARRAYofARRAY8 (&clnt_addr);
@@ -994,7 +1049,7 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
     }
     
     /* Read authentication data */
-    if (! XdmcpReadARRAY8 (&buf, &clnt_authdata)) {
+    if G_UNLIKELY (! XdmcpReadARRAY8 (&buf, &clnt_authdata)) {
 	gdm_error (_("%s: Could not read Authentication Data"),
 		   "gdm_xdmcp_handle_request");
 	XdmcpDisposeARRAYofARRAY8 (&clnt_addr);
@@ -1004,7 +1059,7 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
     }
     
     /* Read and select from supported authorization list */
-    if (! XdmcpReadARRAYofARRAY8 (&buf, &clnt_authorization)) {
+    if G_UNLIKELY (! XdmcpReadARRAYofARRAY8 (&buf, &clnt_authorization)) {
 	gdm_error (_("%s: Could not read Authorization List"),
 		   "gdm_xdmcp_handle_request");
 	XdmcpDisposeARRAY8 (&clnt_authdata);
@@ -1016,11 +1071,12 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
     
     /* libXdmcp doesn't terminate strings properly so we cheat and use strncmp() */
     for (i = 0 ; i < clnt_authorization.length ; i++)
-	if (! strncmp (clnt_authorization.data[i].data, "MIT-MAGIC-COOKIE-1", 18))
+	if (clnt_authorization.data[i].length == 18 &&
+	    strncmp (clnt_authorization.data[i].data, "MIT-MAGIC-COOKIE-1", 18) == 0)
 	    mitauth = TRUE;
     
     /* Manufacturer ID */
-    if (! XdmcpReadARRAY8 (&buf, &clnt_manufacturer)) {
+    if G_UNLIKELY (! XdmcpReadARRAY8 (&buf, &clnt_manufacturer)) {
 	gdm_error (_("%s: Could not read Manufacturer ID"),
 		   "gdm_xdmcp_handle_request");
 	XdmcpDisposeARRAY8 (&clnt_authname);
@@ -1044,7 +1100,7 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
 	explen += 2+clnt_authorization.data[i].length;
     explen += 2+clnt_manufacturer.length;
     
-    if (explen != len) {
+    if G_UNLIKELY (explen != len) {
 	gdm_error (_("%s: Failed checksum from %s"),
 		   "gdm_xdmcp_handle_request",
 		   inet_ntoa (clnt_sa->sin_addr));
@@ -1056,10 +1112,13 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
 	XdmcpDisposeARRAY16 (&clnt_conntyp);
 	return;
     }
-    
-    gdm_debug ("gdm_xdmcp_handle_request: xdmcp_pending=%d, MaxPending=%d, xdmcp_sessions=%d, MaxSessions=%d",
-	       xdmcp_pending, GdmMaxPending, xdmcp_sessions, GdmMaxSessions);
 
+    if G_UNLIKELY (GdmDebug) {
+	    char *s = g_strndup (clnt_manufacturer.data, clnt_manufacturer.length);
+	    gdm_debug ("gdm_xdmcp_handle_request: xdmcp_pending=%d, MaxPending=%d, xdmcp_sessions=%d, MaxSessions=%d, ManufacturerID=%s",
+		       xdmcp_pending, GdmMaxPending, xdmcp_sessions, GdmMaxSessions, ve_sure_string (s));
+	    g_free (s);
+    }
 
     /* Check if ok to manage display */
     if (mitauth &&
@@ -1226,28 +1285,33 @@ gdm_xdmcp_handle_manage (struct sockaddr_in *clnt_sa, gint len)
     }
     
     /* SessionID */
-    if (! XdmcpReadCARD32 (&buf, &clnt_sessid)) {
+    if G_UNLIKELY (! XdmcpReadCARD32 (&buf, &clnt_sessid)) {
 	gdm_error (_("%s: Could not read Session ID"),
 		   "gdm_xdmcp_handle_manage");
 	return;
     }
     
     /* Remote display number */
-    if (! XdmcpReadCARD16 (&buf, &clnt_dspnum)) {
+    if G_UNLIKELY (! XdmcpReadCARD16 (&buf, &clnt_dspnum)) {
 	gdm_error (_("%s: Could not read Display Number"),
 		   "gdm_xdmcp_handle_manage");
 	return;
     }
     
-    gdm_debug ("gdm_xdmcp_handle_manage: Got Display=%d, SessionID=%ld from %s", 
-	       (int)clnt_dspnum, (long)clnt_sessid, inet_ntoa (clnt_sa->sin_addr));
-    
     /* Display Class */
-    if (! XdmcpReadARRAY8 (&buf, &clnt_dspclass)) {
+    if G_UNLIKELY (! XdmcpReadARRAY8 (&buf, &clnt_dspclass)) {
 	gdm_error (_("%s: Could not read Display Class"),
 		   "gdm_xdmcp_handle_manage");
 	return;
     }
+
+    if G_UNLIKELY (GdmDebug) {
+	    char *s = g_strndup (clnt_dspclass.data, clnt_dspclass.length);
+	    gdm_debug ("gdm_xdmcp_handle_manage: Got Display=%d, SessionID=%ld Class=%s from %s", 
+		       (int)clnt_dspnum, (long)clnt_sessid, ve_sure_string (s), inet_ntoa (clnt_sa->sin_addr));
+	    g_free (s);
+    }
+    
     
     d = gdm_xdmcp_display_lookup (clnt_sessid);
     if (d != NULL &&
@@ -1289,13 +1353,13 @@ gdm_xdmcp_handle_manage (struct sockaddr_in *clnt_sa, gint len)
 	xdmcp_pending--;
 
 	/* Start greeter/session */
-	if (!gdm_display_manage (d)) {
+	if G_UNLIKELY (!gdm_display_manage (d)) {
 	    gdm_xdmcp_send_failed (clnt_sa, clnt_sessid);
 	    XdmcpDisposeARRAY8(&clnt_dspclass);
 	    return;
 	}
     }
-    else if (d != NULL && d->dispstat == XDMCP_MANAGED) {
+    else if G_UNLIKELY (d != NULL && d->dispstat == XDMCP_MANAGED) {
 	gdm_debug ("gdm_xdmcp_handle_manage: Session id %ld already managed",
 		   (long)clnt_sessid);	
     }
@@ -1318,14 +1382,22 @@ gdm_xdmcp_handle_managed_forward (struct sockaddr_in *clnt_sa, gint len)
 		   "Got MANAGED_FORWARD from %s",
 		   inet_ntoa (clnt_sa->sin_addr));
 
+	/* Check with tcp_wrappers if client is allowed to access */
+	if (! gdm_xdmcp_host_allow (clnt_sa)) {
+		gdm_error ("%s: Got MANAGED_FORWARD from banned host %s", 
+			   "gdm_xdmcp_handle_request",
+			   inet_ntoa (clnt_sa->sin_addr));
+		return;
+	}
+
 	/* Hostname */
-	if ( ! XdmcpReadARRAY8 (&buf, &clnt_address)) {
+	if G_UNLIKELY ( ! XdmcpReadARRAY8 (&buf, &clnt_address)) {
 		gdm_error (_("%s: Could not read address"),
 			   "gdm_xdmcp_handle_managed_forward");
 		return;
 	}
 
-	if (clnt_address.length != sizeof (struct in_addr)) {
+	if G_UNLIKELY (clnt_address.length != sizeof (struct in_addr)) {
 		gdm_error (_("%s: Could not read address"),
 			   "gdm_xdmcp_handle_managed_forward");
 		XdmcpDisposeARRAY8 (&clnt_address);
@@ -1350,14 +1422,14 @@ static void
 gdm_xdmcp_whack_queued_managed_forwards (struct sockaddr_in *clnt_sa,
 					 struct in_addr *origin)
 {
-	GList *li;
+	GSList *li;
 
 	for (li = managed_forwards; li != NULL; li = li->next) {
 		ManagedForward *mf = li->data;
 		if (mf->manager.sin_addr.s_addr == clnt_sa->sin_addr.s_addr &&
 		    mf->origin.sin_addr.s_addr == origin->s_addr) {
-			managed_forwards = g_list_remove_link (managed_forwards, li);
-			g_list_free_1 (li);
+			managed_forwards = g_slist_remove_link (managed_forwards, li);
+			g_slist_free_1 (li);
 			g_source_remove (mf->handler);
 			/* mf freed by glib */
 			return;
@@ -1371,17 +1443,25 @@ gdm_xdmcp_handle_got_managed_forward (struct sockaddr_in *clnt_sa, gint len)
 	ARRAY8 clnt_address;
 
 	gdm_debug ("gdm_xdmcp_handle_got_managed_forward: "
-		   "Got MANAGED_FORWARD from %s",
+		   "Got GOT_MANAGED_FORWARD from %s",
 		   inet_ntoa (clnt_sa->sin_addr));
 
+	/* Check with tcp_wrappers if client is allowed to access */
+	if (! gdm_xdmcp_host_allow (clnt_sa)) {
+		gdm_error ("%s: Got GOT_MANAGED_FORWARD from banned host %s", 
+			   "gdm_xdmcp_handle_request",
+			   inet_ntoa (clnt_sa->sin_addr));
+		return;
+	}
+
 	/* Hostname */
-	if ( ! XdmcpReadARRAY8 (&buf, &clnt_address)) {
+	if G_UNLIKELY ( ! XdmcpReadARRAY8 (&buf, &clnt_address)) {
 		gdm_error (_("%s: Could not read address"),
 			   "gdm_xdmcp_handle_got_managed_forward");
 		return;
 	}
 
-	if (clnt_address.length != sizeof (struct in_addr)) {
+	if G_UNLIKELY (clnt_address.length != sizeof (struct in_addr)) {
 		gdm_error (_("%s: Could not read address"),
 			   "gdm_xdmcp_handle_got_managed_forward");
 		XdmcpDisposeARRAY8 (&clnt_address);
@@ -1455,7 +1535,7 @@ gdm_xdmcp_handle_keepalive (struct sockaddr_in *clnt_sa, gint len)
     
     gdm_debug ("gdm_xdmcp_handle_keepalive: Got KEEPALIVE from %s", 
 	       inet_ntoa (clnt_sa->sin_addr));
-    
+
     /* Check with tcp_wrappers if client is allowed to access */
     if (! gdm_xdmcp_host_allow (clnt_sa)) {
 	gdm_error (_("%s: Got KEEPALIVE from banned host %s"), 
@@ -1465,37 +1545,51 @@ gdm_xdmcp_handle_keepalive (struct sockaddr_in *clnt_sa, gint len)
     }
     
     /* Remote display number */
-    if (! XdmcpReadCARD16 (&buf, &clnt_dspnum)) {
+    if G_UNLIKELY (! XdmcpReadCARD16 (&buf, &clnt_dspnum)) {
 	gdm_error (_("%s: Could not read Display Number"),
 		   "gdm_xdmcp_handle_keepalive");
 	return;
     }
     
     /* SessionID */
-    if (! XdmcpReadCARD32 (&buf, &clnt_sessid)) {
+    if G_UNLIKELY (! XdmcpReadCARD32 (&buf, &clnt_sessid)) {
 	gdm_error (_("%s: Could not read Session ID"),
 		   "gdm_xdmcp_handle_keepalive");
 	return;
     }
     
-    gdm_xdmcp_send_alive (clnt_sa, clnt_sessid);
+    gdm_xdmcp_send_alive (clnt_sa, clnt_dspnum, clnt_sessid);
 }
 
 
 static void
-gdm_xdmcp_send_alive (struct sockaddr_in *clnt_sa, CARD32 sessid)
+gdm_xdmcp_send_alive (struct sockaddr_in *clnt_sa, CARD16 dspnum, CARD32 sessid)
 {
     XdmcpHeader header;
+    GdmDisplay *d;
+    int send_running = 0;
+    CARD32 send_sessid = 0;
     
-    gdm_debug ("Sending ALIVE to %ld", (long)sessid);
+    d = gdm_xdmcp_display_lookup (sessid);
+    if (d == NULL)
+	    d = gdm_xdmcp_display_lookup_by_host (&(clnt_sa->sin_addr), dspnum);
+
+    if (d != NULL) {
+	    send_sessid = d->sessionid;
+	    if (d->dispstat == XDMCP_MANAGED)
+		    send_running = 1;
+    }
+
+    gdm_debug ("Sending ALIVE to %ld (running %d, sessid %ld)",
+	       (long)sessid, send_running, (long)send_sessid);
     
     header.version = XDM_PROTOCOL_VERSION;
     header.opcode = (CARD16) ALIVE;
     header.length = 5;
     
     XdmcpWriteHeader (&buf, &header);
-    XdmcpWriteCARD8 (&buf, 1);
-    XdmcpWriteCARD32 (&buf, sessid);
+    XdmcpWriteCARD8 (&buf, send_running);
+    XdmcpWriteCARD32 (&buf, send_sessid);
     XdmcpFlush (gdm_xdmcpfd, &buf, (XdmcpNetaddr)clnt_sa,
 		(int)sizeof (struct sockaddr_in));
 }
@@ -1556,6 +1650,8 @@ gdm_xdmcp_display_alloc (struct in_addr *addr,
     d->console = FALSE;
     d->dispstat = XDMCP_PENDING;
     d->sessionid = globsessid++;
+    if (d->sessionid == 0)
+	    d->sessionid = globsessid++;
     d->acctime = time (NULL);
     d->dispnum = displaynum;
 
@@ -1592,9 +1688,13 @@ gdm_xdmcp_display_alloc (struct in_addr *addr,
 
     d->slave_notify_fd = -1;
     d->master_notify_fd = -1;
+
+    d->xsession_errors_bytes = 0;
+    d->xsession_errors_fd = -1;
+    d->session_output_fd = -1;
     
     /* Secure display with cookie */
-    if (! gdm_auth_secure_display (d))
+    if G_UNLIKELY (! gdm_auth_secure_display (d))
 	gdm_error ("gdm_xdmcp_display_alloc: Error setting up cookies for %s", d->name);
     
     displays = g_slist_append (displays, d);
@@ -1666,6 +1766,7 @@ static void
 gdm_xdmcp_displays_check (void)
 {
 	GSList *dlist;
+	time_t curtime = time (NULL);
 
 	dlist = displays;
 	while (dlist != NULL) {
@@ -1674,7 +1775,7 @@ gdm_xdmcp_displays_check (void)
 		if (d != NULL &&
 		    d->type == TYPE_XDMCP &&
 		    d->dispstat == XDMCP_PENDING &&
-		    time (NULL) > d->acctime + GdmMaxManageWait) {
+		    curtime > d->acctime + GdmMaxManageWait) {
 			gdm_debug ("gdm_xdmcp_displays_check: Disposing session id %ld",
 				   (long)d->sessionid);
 			gdm_display_dispose (d);
