@@ -416,6 +416,7 @@ setup_user_combo (const char *name, const char *key)
 	struct passwd *pwent;
 	char *str;
 	int cnt;
+	time_t time_started;
 
 	str = ve_config_get_string (ve_config_get (GDM_CONFIG_FILE), key);
 
@@ -425,6 +426,8 @@ setup_user_combo (const char *name, const char *key)
 	if ( ! ve_string_empty (str))
 		users = g_list_append (users, g_strdup (str));
 
+	time_started = time (NULL);
+
 	setpwent ();
 
 	pwent = getpwent();
@@ -437,7 +440,8 @@ setup_user_combo (const char *name, const char *key)
 			users = g_list_append (users,
 					       g_strdup (pwent->pw_name));
 			/* FIXME: fix properly, see bug #111830 */
-			if (cnt >= 50) {
+			if (cnt >= 100 ||
+			    time_started + 5 <= time (NULL)) {
 				users = g_list_append (users,
 						       g_strdup (_("Too many users to list here...")));
 				break;
@@ -1977,15 +1981,25 @@ dialog_response (GtkWidget *dlg, int response, gpointer data)
 		timeout_remove_all ();
 		gtk_main_quit ();
 	} else if (response == GTK_RESPONSE_HELP) {
-		/* FIXME: when help gets written connect it here */
+		GError *error = NULL;
 		GtkWidget *setup_dialog = glade_helper_get
 			(xml, "setup_dialog", GTK_TYPE_WINDOW);
-		static GtkWidget *dlg;
+		static GtkWidget *dlg = NULL;
 
 		if (dlg != NULL) {
 			gtk_window_present (GTK_WINDOW (dlg));
 			return;
 		}
+
+		if ( ! RUNNING_UNDER_GDM) {
+			gnome_help_display_uri ("ghelp:gdm", &error);
+			/* FIXME: handle errors nicer */
+			if (error == NULL)
+				return;
+			g_error_free (error);
+		}
+
+		/* fallback help dialogue */
 	
 		/* HIG compliance? */
 		dlg = gtk_message_dialog_new
@@ -2093,6 +2107,12 @@ setup_gui (void)
 	add_to_size_group (sg, "timedlogin");
 	g_object_unref (G_OBJECT (sg));
 
+	/* Note: Just setting up a notify here doesn't mean
+	   things will get updated.  You must also wire in
+	   the logic for update in the daemon and/or greeters
+	   (or chooser) or whereever this is actually being
+	   updated. */
+
 	setup_user_combo ("autologin_combo",
 			  GDM_KEY_AUTOMATICLOGIN);
 	setup_user_combo ("timedlogin_combo",
@@ -2318,6 +2338,7 @@ main (int argc, char *argv[])
 	gnome_program_init ("gdmsetup", VERSION, 
 			    LIBGNOMEUI_MODULE /* module_info */,
 			    argc, argv,
+			    GNOME_PROGRAM_STANDARD_PROPERTIES,
 			    /* *GNOME_PARAM_POPT_TABLE, options, */
 			    GNOME_PARAM_CREATE_DIRECTORIES, ! RUNNING_UNDER_GDM,
 			    NULL);
